// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.js; 

import org.ibex.util.*; 
import java.io.*;
import java.util.*;

/** The minimum set of functionality required for objects which are manipulated by JavaScript */
public interface JS extends Pausable {

    /** Returns an enumeration of the keys in this object. */
    public JS.Enumeration keys() throws JSExn;

    /** Return the value associated with the given key. */
    public JS get(JS key) throws JSExn;

    /** Store a specific value against the given key. */
    public void put(JS key, JS value) throws JSExn;

    /** Executes or unpauses the task, running it in a
     *  <i>pausable</i> context. */
    public Object run(Object o) throws Exception, AlreadyRunningException;

    /** Pauses the running task at its convienience. */
    public void pause() throws NotPausableException;

    /** Calls a specific method with given arguments on this object.
     *  An exception is thrown if there is no such method or the
     *  arguments are invalid. */
    public JS call(JS method, JS[] args) throws JSExn;

    /** Calls this object with the given arguments, running it in an
     *  <i>unpausable</i> context. An exception is thrown if this
     *  object is not callable or the arguments are invalid. */
    public JS call(JS[] args) throws JSExn;

    /** Returns the names of the formal arguments, if any.
     *  This function must never return a null object. */
    public String[] getFormalArgs();

    /** Returns true if the specific key is found in this object. */
    public boolean hasKey(JS key);

    /** Put a value to the given key, calling any write traps that have
     *  been placed on the key.
     *
     *  This function may block for an indeterminate amount of time.
     *
     *  Write traps may change the final value  before it is put, thus
     *  be careful to read the latest value from this function's return
     *  before doing any more work with it.
     */
    public JS putAndTriggerTraps(JS key, JS value) throws JSExn;

    /** Gets the value for a given key, calling any read traps that have
     *  been placed on the key.
     *
     *  This function may block for an indeterminate amount of time.
     */
    public JS getAndTriggerTraps(JS key) throws JSExn;

    /** Calls the write traps placed on a given key with the given value
     *  and returns the result without saving it to this object's key store. */
    public JS justTriggerTraps(JS key, JS value) throws JSExn;

    public void addTrap(JS key, JS function) throws JSExn;
    public void delTrap(JS key, JS function) throws JSExn;
    public Trap getTrap(JS key) throws JSExn;

    // FIXME: consider renaming/removing these
    public InputStream getInputStream() throws IOException, JSExn;
    public JS unclone();
    public String coerceToString();


    // Implementations ////////////////////////////////////////////////////////

    /** Provides the lightest possible implementation of JS, throwing
     *  exceptions on every mutable operation. */
    public static class Immutable implements JS {
        private static final String[] emptystr = new String[0];

        public JS unclone() { return this; }
        public JS.Enumeration keys() throws JSExn { throw new JSExn(
            "object has no key set, class ["+ getClass().getName() +"]"); }
        public JS get(JS key) throws JSExn { return null; }
        public void put(JS key, JS val) throws JSExn { throw new JSExn(
            "'" + key + "' is read only on class ["+ getClass().getName() +"]"); }
        public InputStream getInputStream() throws IOException, JSExn { throw new JSExn(
            "object has not associated stream, class ["+ getClass().getName() +"]"); }

        public boolean hasKey(JS key) { return false; }

        public Object run(Object o) throws Exception { throw new JSExn(
            "object cannot be called, class ["+ getClass().getName() +"]"); }
        public void pause() { throw new NotPausableException(); }

        public JS call(JS[] args) throws JSExn { throw new JSExn(
            "object cannot be called, class ["+ getClass().getName() +"]"); }
        public JS call(JS method, JS[] args) throws JSExn { throw new JSExn(
            "method not found: " + Script.str(method)); }
        public String[] getFormalArgs() { return emptystr; }

        public void declare(JS key) throws JSExn { throw new JSExn(
            "object cannot declare key: "+ Script.str(key)); }
        public void undeclare(JS key) throws JSExn { } // FIXME throw error?

        public JS putAndTriggerTraps(JS key, JS val) throws JSExn { throw new JSExn(
            "'" + key + "' is trap read only on class ["+ getClass().getName() +"]"); }
        public JS getAndTriggerTraps(JS key) throws JSExn { return null; } // FIXME throw errors?
        public JS justTriggerTraps(JS key, JS value) throws JSExn { return null; }

        public void addTrap(JS key, JS function) throws JSExn { throw new JSExn(
            "'" + key + "' is not trappable on class ["+ getClass().getName() +"]"); }
        public void delTrap(JS key, JS function) throws JSExn { throw new JSExn(
            "'" + key + "' trap is read only on class ["+ getClass().getName() +"]"); }
        public Trap getTrap(JS key) throws JSExn { throw new JSExn(
            "'" + key + "' is not trappable on class ["+ getClass().getName() +"]"); }

        public String coerceToString() { return "object"; }
    }

    public interface Cloneable {}

    public static class Clone implements JS {
        protected final JS clonee;
        public Clone(JS clonee) throws JSExn {
            if (!(clonee instanceof Cloneable)) throw new JSExn(
                clonee.getClass().getName() + " is not implement cloneable");
            this.clonee = clonee;
        }

        public JS unclone() { return clonee.unclone(); }
        public boolean equals(Object o) { return clonee.equals(o); }

        public Enumeration keys() throws JSExn { return clonee.keys(); }
        public JS get(JS k) throws JSExn { return clonee.get(k); }
        public void put(JS k, JS v) throws JSExn { clonee.put(k, v); }
        public InputStream getInputStream() throws IOException, JSExn {
            return clonee.getInputStream(); }

        public Object run(Object o) throws Exception { return clonee.run(o); }
        public void pause() { clonee.pause(); }

        public JS call(JS m, JS[] a) throws JSExn { return clonee.call(m, a); }
        public JS call(JS[] a) throws JSExn { return clonee.call(a); }
        public String[] getFormalArgs() { return clonee.getFormalArgs(); }

        public boolean hasKey(JS k) { return clonee.hasKey(k); }

        public JS putAndTriggerTraps(JS k, JS v) throws JSExn {
            return clonee.putAndTriggerTraps(k, v); }
        public JS getAndTriggerTraps(JS k) throws JSExn {
            return clonee.getAndTriggerTraps(k); }
        public JS justTriggerTraps(JS k, JS v) throws JSExn {
            return clonee.justTriggerTraps(k, v); }

        public void addTrap(JS k, JS f) throws JSExn { clonee.addTrap(k, f); }
        public void delTrap(JS k, JS f) throws JSExn { clonee.delTrap(k, f); }
        public Trap getTrap(JS k) throws JSExn { return clonee.getTrap(k); }

        public String coerceToString() { return clonee.coerceToString(); }
    }

    public static class Obj extends Basket.Hash implements JS {
        private static final String[] emptystr = new String[0];
        private static final Placeholder holder = new Placeholder();

        /** entries[index + 0] // key
         *  entries[index + 1] // value
         *  entries[index + 2] // trap
         */
        protected final int indexmultiple = 3;

        protected void entryAdded(int p) {}
        protected void entryUpdated(int p) {}
        protected void entryRemoved(int p) {}

        public Obj() { super(4, 0.75F); }

        public JS unclone() { return this; }
        public InputStream getInputStream() throws IOException, JSExn { throw new JSExn(
            "object has not associated stream, class ["+ getClass().getName() +"]"); }

        public Object run(Object o) throws Exception { throw new JSExn(
            "object cannot be called, class ["+ getClass().getName() +"]"); }
        public void pause() { throw new NotPausableException(); }

        public JS call(JS[] args) throws JSExn { throw new JSExn(
            "object cannot be called, class ["+ getClass().getName() +"]"); }
        public JS call(JS method, JS[] args) throws JSExn { throw new JSExn(
            "method not found: " + Script.str(method)); }
        public String[] getFormalArgs() { return emptystr; }

        public Enumeration keys() throws JSExn {
            return new Enumeration(null) {
                private int dest = -1, next = -1;
                public boolean _hasNext() {
                    for (int i = Math.max(0, dest); i < usedslots; i++)
                        if (i > 0 ? entries[i * indexmultiple] != null : true &&
                            entries[i * indexmultiple] != this) { next = i; return true; }
                    return false;
                }
                public JS _next() throws JSExn {
                    if (next < 0 && !hasNext()) throw new NoSuchElementException();
                    int index = next; dest = next; next = -1;
                    return (JS)entries[index * indexmultiple];
                }
            };
        }
        public JS get(JS key) throws JSExn { int i = indexOf(key);
            return i < 0 ? null : entries[i + 1] instanceof Placeholder ?  null : (JS)entries[i + 1]; }
        public void put(JS key, JS val) throws JSExn {
            // NOTE: only way value can be stored as null is using declare()
            int dest = put(indexOf(key), key);
            if (val == null) entries[dest + 1] = holder;
            else entries[dest + 1] = val; }

        public boolean hasKey(JS key) { return indexOf(key) >= 0; }
        /*public boolean hasValue(JS key, JS value) {
            int i = indexOf(key); return i >= 0 && entries[i + 1] != null; }
        public boolean hasTrap(JS key, JS trap) {
            int i = indexOf(key); return i >= 0 && entries[i + 2] != null; }*/

        public void declare(JS key) throws JSExn { entries[put(indexOf(key), key) + 1] = null; }
        public void undeclare(JS key) throws JSExn { remove(indexOf(key)); }

        public JS putAndTriggerTraps(JS key, JS val) throws JSExn {
            Trap t = null; int i = indexOf(key);
            if (i >= 0) t = (Trap)entries[i + 2];
            if (t != null && (t = t.write()) != null) {
                val = (JS)new Interpreter(t, val, false).run(null);
            }
            if (val == null) entries[i + 1] = holder;
            else entries[i + 1] = val;
            return val;
        }
        public JS getAndTriggerTraps(JS key) throws JSExn {
            Trap t = null; int i = indexOf(key);
            if (i < 0) return null;
            t = (Trap)entries[i + 2];
            return t == null ? (JS)entries[i + 1] : (JS)new Interpreter(t, null, false).run(null);
        }
        public JS justTriggerTraps(JS key, JS val) throws JSExn {
            Trap t = null; int i = indexOf(key);
            if (i >= 0) t = (Trap)entries[i + 2];
            if (t == null || (t = t.write()) == null) return val;
            return (JS)new Interpreter(t, val, true).run(null);
        }

        public void addTrap(JS key, JS f) throws JSExn {
            if (f.getFormalArgs() == null || f.getFormalArgs().length > 1) throw new JSExn(
                "traps must take either one argument (write) or no arguments (read)");
            int i = indexOf(key); if (i < 0) i = put(i, key);
            for (Trap t = (Trap)entries[i + 2]; t != null; t = t.next())
                if (t.function().equals(f)) return;
            entries[i + 2] = new TrapHolder(this, key, f, (Trap)entries[i + 2]);
        }

        public void delTrap(JS key, JS f) throws JSExn {
            int i = indexOf(key); if (i < 0) return;
            Trap t = (Trap)entries[i + 2];
            if (t.function().equals(f)) { entries[i + 2] = t.next(); return; }
            for (; t.next() != null; t = t.next())
                if (t.next().function().equals(f)) { ((TrapHolder)t).next = t.next().next(); return; }
        }

        public Trap getTrap(JS key) throws JSExn {
            int i = indexOf(key); return i < 0 ? null : (Trap)entries[i + 2];
        }

        public String coerceToString() { return "object"; }

        private static final class Placeholder implements Serializable {}

        private static final class TrapHolder implements Trap {
            private final JS target, key, function;
            private Trap next;
            TrapHolder(JS t, JS k, JS f, Trap n) { target = t; key = k; function = f; next = n; }

            public JS key() { return key; }
            public JS target() { return target; }
            public JS function() { return function; }

            public boolean isReadTrap()  {
                return function.getFormalArgs() == null || function.getFormalArgs().length == 0; }
            public boolean isWriteTrap() {
                return function.getFormalArgs() != null && function.getFormalArgs().length != 0; }

            public Trap next() { return next; }
            public Trap nextRead() {
                Trap t = next; while (t != null && t.isWriteTrap()) t = t.next(); return t; }
            public Trap nextWrite() {
                Trap t = next; while (t != null && t.isReadTrap()) t = t.next(); return t; }
            public Trap read() { return isReadTrap() ? this : nextRead(); }
            public Trap write() { return isWriteTrap() ? this : nextWrite(); }
        }
    }

    public interface Trap {
        public JS key();
        public JS target();
        public JS function();
        public boolean isReadTrap();
        public boolean isWriteTrap();
        public Trap next();
        public Trap nextRead();
        public Trap nextWrite();

        public Trap read(); // FIXME reconsider these function names
        public Trap write();
    }

    /** An empty class used for instanceof matching the result of JS.get()
     *  to decide if the key is a method (to be called with JS.callMethod(). */
    public static final class Method extends Immutable {}


    public static abstract class Enumeration extends Immutable {
        public static final Enumeration EMPTY = new Empty(null);

        private final Enumeration parent;

        public Enumeration(Enumeration parent) { this.parent = parent; }

        protected abstract boolean _hasNext();
        protected abstract JS _next() throws JSExn;

        public final boolean hasNext() {
            return _hasNext() || parent != null ? parent.hasNext() : false; }
        public final JS next() throws JSExn {
            if (_hasNext()) return _next();
            if (parent == null) throw new NoSuchElementException("reached end of set");
            return parent.next();
        }

        public JS get(JS key) throws JSExn {
            //#switch(Script.str(key))
            case "hasNext": return Script.B(hasNext());
            case "next": return next();
            //#end
            return super.get(key);
        }

        public static final class Empty extends Enumeration {
            public Empty(Enumeration parent) { super(parent); }
            protected boolean _hasNext() { return false; }
            protected JS _next() { throw new NoSuchElementException(); }
        }

        public static final class JavaIterator extends Enumeration {
            private final Iterator i;
            public JavaIterator(Enumeration parent, Iterator i) { super(parent); this.i = i; }
            protected boolean _hasNext() { return i.hasNext(); }
            protected JS _next() { return (JS)i.next(); }
        }

        public static final class RandomAccessList extends Enumeration {
            private final List l;
            private int pos = 0, size;
            public RandomAccessList(Enumeration parent, List list) {
                super(parent); l = list; size = l.size(); }
            protected boolean _hasNext() { return  pos < size; }
            protected JS _next() { return (JS)l.get(pos++); }
        }
    }

}
